from __future__ import absolute_import

import collections
import urllib

import requests

import tornado.web
import tornado.httpclient

try:
    from tornado.curl_httpclient import CurlAsyncHTTPClient as AsyncHTTPClient
except ImportError:
    from tornado.httpclient import AsyncHTTPClient

try:
    import ujson as json
except ImportError:
    import json


#region Internal


http_client = AsyncHTTPClient(max_clients=1024)
http_session = requests.session()


class NoneObject(object):
    __slots__ = {}

    def __getattribute__(self, item):
        return self

    def __getitem__(self, item):
        return self

    def __nonzero__(self):
        return False

    def __contains__(self, item):
        return False

    def __iter__(self):
        return iter([])

    def __repr__(self):
        return ""

    def __str__(self):
        return ""


class Object(object):
    _none = NoneObject()

    @staticmethod
    def create(d):
        if isinstance(d, dict):
            return Object(d)
        elif isinstance(d, (set, list, tuple)):
            return [Object.create(i) for i in d]
        else:
            return d

    def __init__(self, d):
        for k, v in d.iteritems():
            if isinstance(v, dict):
                self.__dict__[k] = Object(v)
            elif isinstance(v, (set, list, tuple)):
                self.__dict__[k] = [Object.create(i) for i in v]
            else:
                self.__dict__[k] = v

    def __getattr__(self, name):
        return self.__dict__.get(name, Object._none)

    def __getitem__(self, name):
        return self.__dict__.get(name, Object._none)

    def __iter__(self):
        return self.__dict__.itervalues()

    def __repr__(self):
        return str(self.__dict__)


#endregion


class Client(object):
    class BaseClient(object):
        def __init__(self, client):
            self.client = client

    class Item(BaseClient):
        def get(self, item_id, language=None, callback=None):
            return self.client.get("item/get", item_id=item_id, language=language, callback=callback)

        def search(self, category_id=None, entity_id=None, query=None, brands=None, skip=0, take=10, language=None,
                   sort=None, callback=None):
            return self.client.get("item/search", category_id=category_id, entity_id=entity_id, query=query,
                                   _brands=brands,
                                   skip=skip, take=take, language=language, sort=sort, callback=callback)

        def list(self, item_ids, language=None, callback=None):
            return self.client.get("item/list", _item_ids=item_ids, language=language, callback=callback)

    class News(BaseClient):
        def get(self, article_id, language, callback=None):
            pass

        def search(self, skip=0, take=20, language="ru", callback=None):
            return self.client.get("news/search", skip=skip, take=take, language=language, callback=callback)

    class Customer(BaseClient):
        def get_by_claimed_id(self, origin, claimed_id, callback=None):
            return self.client.get("customer/get_by_claimed_id", origin=origin, claimed_id=claimed_id,
                                   callback=callback)

        def get_by_email(self, email, callback=None):
            return self.client.get("customer/get_by_email", email=email, callback=callback)

        def register(self, email, name, surname, birthdate=None, avatar=None, origin=None, claimed_id=None,
                     callback=None):
            return self.client.get("customer/register", email=email, name=name, surname=surname, birthdate=birthdate,
                                   avatar=avatar, origin=origin, claimed_id=claimed_id, callback=callback)

        def connect(self, customer_id, origin, claimed_id, callback=None):
            return self.client.get("customer/connect", customer_id=customer_id, origin=origin, claimed_id=claimed_id,
                                   callback=callback)

    class Category(BaseClient):
        def get(self, category_id, language, callback=None):
            return self.client.get("category/get", category_id=category_id, language=language, callback=callback)

        def tree(self, language, callback=None):
            return self.client.get("category/tree", language=language, callback=callback)

    class Entity(BaseClient):
        def get(self, entity_id, language, callback=None):
            return self.client.get("entity/get", entity_id=entity_id, language=language, callback=callback)

    class Email(BaseClient):
        def send(self, from_email, from_name, to, subject, message, callback=None):
            return self.client.get("email/send", from_email=from_email, from_name=from_name, to=to,
                                   subject=subject, message=message, callback=callback)

    class Currency(BaseClient):
        def rates(self, currency, callback=None):
            return self.client.get("currency/rates", currency=currency, callback=callback)

    def __init__(self, url):
        self.url = url.strip("/")

        self.item = Client.Item(self)
        self.news = Client.News(self)
        self.customer = Client.Customer(self)
        self.category = Client.Category(self)
        self.entity = Client.Entity(self)
        self.email = Client.Email(self)
        self.currency = Client.Currency(self)

    def request(self, verb, path, data=None, callback=None):
        url = "%s/%s" % (self.url, path)
        if verb == "GET":
            #TODO: review
            url += "?%s" % urllib.urlencode(dict((k, v.encode("utf8") if isinstance(v, unicode) else v) for k, v in data.iteritems() if v is not None), True)
            data = None
        else:
            if data:
                data = json.dumps(data)

        response = http_session.request(verb, url, data=data)
        status_code = response.status_code
        if 200 == status_code:
            return Object.create(json.loads(response.content))
        elif 404 == status_code:
            return None
        else:
            raise requests.HTTPError(response.status_code)

    def get(self, path, callback, **query_args):
        return self.request("GET", path, query_args, callback)

    def pipeline(self):
        return ClientPipeline(self)


class ClientPipeline(Client):
    def __init__(self, client):
        self._client = client
        self._pipe = []

        super(ClientPipeline, self).__init__(client.url)

    def request(self, verb, path, data=None, callback=None):
        self._pipe.append({"path": path, "args": data})

    def dummy(self):
        self._pipe.append(None)

    def execute(self, callback):
        self._client.request("POST", "", self._pipe, callback)

    def pipeline(self):
        raise Exception("Cannot create pipeline in pipeline")


class TornadoClient(Client):
    def request(self, verb, path, data=None, callback=None):
        def _callback(response):
            if response.code == 200:
                callback(Object.create(json.loads(response.body)) if response.body else None)
            elif response.code == 404:
                callback(None)
            else:
                raise tornado.httpclient.HTTPError(response.code, response.effective_url)

        url = "%s/%s" % (self.url, path)
        if verb == "GET":
            #TODO: review
            url += "?%s" % urllib.urlencode(dict((k, v.encode("utf8") if isinstance(v, unicode) else v) for k, v in data.iteritems() if v is not None), True)
            data = None
        else:
            if data:
                data = json.dumps(data)

        http_client.fetch(tornado.httpclient.HTTPRequest(url, method=verb, body=data), _callback)